import enum
import typing
from . import types


class MessageType(enum.IntEnum):
    INVALID = 0
    METHOD_CALL = 1
    METHOD_RETURN = 2
    ERROR = 3
    SIGNAL = 4


class HeaderField(enum.IntEnum):
    INVALID = 0
    PATH = 1
    INTERFACE = 2
    MEMBER = 3
    ERROR_NAME = 4
    REPLY_SERIAL = 5
    DESTINATION = 6
    SENDER = 7
    SIGNATURE = 8
    UNIX_FDS = 9

_next_serial = 0
def next_serial():
    global _next_serial
    _next_serial += 1
    return _next_serial

def _str_field_property(field: HeaderField, cast_to=None):
    def getter(msg):
        ret = msg.headers.get(field, None)
        return ret

    def setter(msg, value):
        if value is None:
            deleter(msg)
        else:
            if cast_to:
                value = cast_to(value)
            msg.headers[field] = value

    def deleter(msg):
        del msg.headers[field]

    return getter, setter, deleter

class Message:
    member = property(*_str_field_property(HeaderField.MEMBER))
    interface = property(*_str_field_property(HeaderField.INTERFACE))
    sender = property(*_str_field_property(HeaderField.SENDER))
    destination = property(*_str_field_property(HeaderField.DESTINATION))
    error_name = property(*_str_field_property(HeaderField.ERROR_NAME))
    path = property(*_str_field_property(HeaderField.PATH, types.ObjectPath))
    signature = property(*_str_field_property(HeaderField.SIGNATURE, types.Signature))

    def __init__(self):
        self.headers = {}                                   # type: typing.Dict[HeaderField, object]
        self.message_type = None                             # type: MessageType
        self.payload = None
        self.serial = 1
        self.flags = 0

    def __str__(self):
        return 'Message(%s, [%s], %s)' % (
            self.message_type,
            ', '.join(['%s=%s' % (k.name, self.headers[k]) for k in sorted(self.headers)]),
            self.payload)

    @property
    def reply_serial(self) -> int:
        v = self.headers[HeaderField.REPLY_SERIAL]
        if isinstance(v, types.enforce_type):
            v = v.value
        return v

    @reply_serial.setter
    def reply_serial(self, v: int):
        if v:
            self.headers[HeaderField.REPLY_SERIAL] = types.enforce_type(v, b'u')
        else:
            del self.reply_serial

    @reply_serial.deleter
    def reply_serial(self):
        del self.headers[HeaderField.REPLY_SERIAL]


def make_mesage(m_type: MessageType,
                bus_name: str,
                interface_name: str,
                member: str,
                object_path: str,
                signature: str=None,
                data=None) -> Message:
    ret = Message()
    ret.message_type = m_type
    if not bus_name is None:
        ret.headers[HeaderField.DESTINATION] = bus_name
    if not interface_name is None:
        ret.headers[HeaderField.INTERFACE] = interface_name
    if signature is None:
        assert data is None
    else:
        ret.headers[HeaderField.SIGNATURE] = types.Signature(signature)
        ret.payload = data
    if not object_path is None:
        ret.headers[HeaderField.PATH] = types.ObjectPath(object_path)
    if not member is None:
        ret.headers[HeaderField.MEMBER] = member
    ret.serial = next_serial()
    return ret


import inspect
from debus.message import Message
import debus.types
import logging
import asyncio
import typing

logger = logging.getLogger(__name__)

class MethodInfo:
    def __init__(self, name, in_signature, out_signature, single_return):
        self.name: str = name
        self.in_signature: str = in_signature
        self.out_signature: str = out_signature
        self.single_return: bool = single_return

class SignalInfo:
    def __init__(self, name, arg_names, signature):
        self.name: str = name
        self.argument_names: typing.List[str] = arg_names
        self.signature: str = signature

class PropertyInfo:
    def __init__(self, name, readonly):
        self.name: str = name
        self.readonly: bool = readonly

def dbus_method(signature: str, return_signature: str, name: str=None, single_return_value=True):
    def dec(foo):
        method_name = name or foo.__name__
        foo._pybus_method = MethodInfo(method_name, signature, return_signature, single_return_value)
        return foo
    return dec


def dbus_signal(signature: str, name:str=None, arg_names:list=None):
    def dec(foo):
        method_name = name or foo.__name__
        _arg_names = arg_names
        if _arg_names is None:
            _arg_names = list(list(inspect.signature(foo).parameters)[1:])
        def _signal_wrapper(_self, *args):
            assert isinstance(_self, DBusInterface)
            _self._obj._bus.emit(_self._obj.path, _self.name, method_name, signature, args)
        _signal_wrapper._pybus_signal = SignalInfo(method_name, _arg_names, signature)
        return _signal_wrapper
    return dec


def dbus_property(signature: str, name:str=None):
    def dec(prop):
        prop._pybus_property = PropertyInfo(name, hasattr(prop, 'fset'))
        return prop
    return dec

class DBusInterface:
    name = None

    def __init__(self):
        self._dbus_methods = {}
        self._dbus_properties = {}
        self._dbus_signals = []
        self._obj: 'DBusObject' = None
        for name, m in inspect.getmembers(self.__class__, lambda x: hasattr(x, '_pybus_method')):
            self._dbus_methods[m._pybus_method.name] = m
        for name, p in inspect.getmembers(self.__class__, lambda x: hasattr(x, '_pybus_property')):
            # properties don't know their names
            p._pybus_property.name = p._pybus_property.name or name
            self._dbus_properties[p._pybus_property.name] = p
        for i in inspect.getmembers(self.__class__, lambda x: hasattr(x, '_pybus_signal')):
            self._dbus_signals.append(i[1])
    def set_object(self, obj: 'DBusObject'):
        self._obj = obj

    def on_method_call(self, msg: Message):
        if msg.member in self._dbus_methods:
            method = self._dbus_methods[msg.member]
            spec: MethodInfo = method._pybus_method
            args = [] if msg.payload is None else msg.payload
            ret = method(self, *args)
            return debus.types.enforce_type(ret, spec.out_signature.encode())
        raise NotImplementedError('Unknown method (%s.)%s, existing methods: %s' % (self.name, msg.member, self._dbus_methods.keys()))

    @property
    def methods(self):
        return list(self._dbus_methods.values())

    @property
    def signals(self):
        return list(self._dbus_signals)

    @property
    def properties(self):
        return list(self._dbus_properties)


class DBusObject:
    def __init__(self, conn, path):
        # type: (debus.ManagedConnection)->None
        self._interfaces = {}
        if isinstance(conn, debus.ClientConnection):
            self._bus = conn
        else:
            self._bus = conn._connection
        self._path = path

    @property
    def path(self):
        return self._path

    @property
    def interfaces(self):
        # type: ()->list[DBusInterface]
        return list(self._interfaces.values())

    def add_interface(self, iface: DBusInterface):
        self._interfaces[iface.name] = iface
        iface.set_object(self)

    def on_method_call(self, msg: Message):
        if msg.interface in self._interfaces:
            return self._interfaces[msg.interface].on_method_call(msg)
        raise NotImplementedError("Interface %r is not supported by object %r" % (msg.interface, self.path))


class ObjectManager:
    def __init__(self, bus):
        # type: (debus.ClientConnection)->None
        # FIXME cyclic dependency :-/
        import debus.freedesktop.introspect
        self._objects = {}
        self._bus = bus
        self.root_object = DBusObject(self._bus, '/')
        self.root_introspect = debus.freedesktop.introspect.IntrospectInterface()
        self.root_object.add_interface(self.root_introspect)
        self.register_object(self.root_object)

    def register_object(self, obj: DBusObject):
        self._objects[obj.path] = obj
        self.root_introspect.child_objects.append(obj.path)

    async def send_return_async(self, method_call: debus.message.Message, ret: debus.types.enforce_type):
        try:
            ret._value = await ret.value
            self.send_return(method_call, ret)
        except Exception as ex:
            logger.exception("Exception during method call %s", method_call)
            self.send_error(method_call, ex)

    def send_error(self, method_call: debus.message.Message, exc: Exception):
        err = debus.Message()
        err.message_type = debus.MessageType.ERROR
        err.destination = method_call.sender
        err.error_name = 'space.equi.debus.Error.%s' % exc.__class__.__name__
        err.reply_serial = method_call.serial
        err.signature = 's'
        err.payload = (repr(exc),)
        logger.warning("Sending an error %s as a response to %s", err, method_call)
        self._bus.send_message(err)

    def send_return(self, method_call: debus.message.Message, ret):
        if isinstance(ret, debus.types.enforce_type):
            signature = ret.signature
            ret = ret._value
        else:
            signature = debus.types.guess_signature(ret)
            logger.warning("Had to guess the signature, got %s for %r", signature, ret)
        ret_msg = debus.Message()
        ret_msg.message_type = debus.MessageType.METHOD_RETURN
        ret_msg.destination = method_call.sender
        ret_msg.reply_serial = method_call.serial
        ret_msg.signature = signature
        ret_msg.payload = ret
        logging.warning("Sending return value %s as a response to %s", ret_msg, method_call)
        self._bus.send_message(ret_msg)

    def handle_call(self, msg: debus.message.Message):
        try:
            path = str(msg.path)
            if path in self._objects:
                ret = self._objects[path].on_method_call(msg)      # type: debus.types.enforce_type
                if isinstance(ret.value, asyncio.Future) or inspect.iscoroutine(ret.value):
                    asyncio.ensure_future(self.send_return_async(msg, ret))
                else:
                    self.send_return(msg, ret)
            else:
                raise NotImplementedError("Unknown path: %s, possible paths: %r" % (path, sorted(self._objects.keys())))
        except Exception as ex:
            logging.exception('Failed to process message %s', msg)
            self.send_error(msg, ex)
            pass




import asyncio
import logging
import debus.types
from debus.wire import WireConnection
import debus.objects
from .message import Message, make_mesage, MessageType, HeaderField
from lxml import etree
import io

try:
    import typing
    if typing.TYPE_CHECKING:
        StringOrBytes = typing.Union[str, bytes]
except:
    pass

logger = logging.getLogger(__name__)

class DBusError(RuntimeError):
    pass


class DBusMethod:
    def __init__(self, obj, iface, name, signature, output):
        # type: (IntrospectedObject, IntrospectedInterface, str, str, str) -> None
        self.obj = obj
        self.iface = iface
        self.name = name
        self.signature = signature
        self.signature_out = output

    def __call__(self, *args, **kwargs):
        return self.obj.bus.call(self.obj.bus_name, self.obj.object_path, self.iface.name, self.name, self.signature, args)

    def __str__(self):
        return '[%s].%s(%r)->%r' % (self.iface.name, self.name, self.signature, self.signature_out)


class IntrospectedSignal:
    def __init__(self, obj, iface, name, signature):
        self.obj = obj
        self.iface = iface
        self.name = name
        self.signature = signature

    def __str__(self):
        return '[%s].%s (%s)' % (self.iface.name, self.name, self.signature)


class IntrospectedInterface:
    def __init__(self, name):
        self.name = name
        self.signals = {}
        self.methods = {}

    def __getattr__(self, item):
        if item in self.methods:
            return self.methods[item]
        raise AttributeError(item)


class IntrospectedObject:
    def __init__(self, bus, bus_name, object_path, introspect_result):
        # type: (ClientConnection, str, str, str) -> None
        self.bus = bus              # type: ClientConnection
        self.bus_name = bus_name
        self.object_path = object_path
        self.interfaces = {}    # type: typing.Dict[str, IntrospectedInterface]
        tree = etree.parse(io.StringIO(introspect_result))
        for interface_node in tree.xpath('/node/interface'):
            interface = IntrospectedInterface(interface_node.get('name'))
            self.interfaces[interface.name] = interface
            for method_node in interface_node.xpath('./method'):
                signature_in = ''.join(method_node.xpath("./arg[@direction='in']/@type"))
                signature_out = ''.join(method_node.xpath("./arg[@direction='out']/@type"))
                method_name = method_node.get('name')
                method_obj = DBusMethod(self, interface, method_name, signature_in, signature_out)
                interface.methods[method_name] = method_obj

            for sig_node in interface_node.xpath('./signal'):
                args = ''.join(sig_node.xpath("./arg/@type"))
                name = sig_node.get('name')
                signal = IntrospectedSignal(self, interface, name, ''.join(args))
                interface.signals[name] = signal

    def log(self, logger):
        logger.warning('DBus object %s on %s:', self.object_path, self.bus_name)
        for ifname in sorted(self.interfaces):
            ifobj = self.interfaces[ifname]
            logger.warning("  Interface %s (%d methods, %d signals)", ifname, len(ifobj.methods), len(ifobj.signals))
            for mname in sorted(ifobj.methods):
                logger.warning("    %s", ifobj.methods[mname])
            for sname in sorted(ifobj.signals):
                logger.warning("    %s", ifobj.signals[sname])


async def get_freedesktop_interface(conn, name=None):
    # type: (debus.connection.ClientConnection, str) -> debus.connection.DBusInterface
    if name:
        name = 'org.freedesktop.DBus.%s' % name
    else:
        name = 'org.freedesktop.DBus'
    return await conn.get_object_interface('org.freedesktop.DBus', '/org/freedesktop/DBus', name)


class ClientConnection:
    def __init__(self, socket=None, uri=None):
        self._wire = WireConnection(socket=socket, uri=uri)
        self._futures = {}       # type: typing.Dict[int, asyncio.Future]
        self._freedesktop_interface = None

    async def connect(self):
        await self._wire.connect_and_auth()
        asyncio.ensure_future(self.run())
        await self.call('org.freedesktop.DBus', '/org/freedesktop/DBus', 'org.freedesktop.DBus', 'Hello', '', None)
        self._freedesktop_interface = await get_freedesktop_interface(self)

    async def run(self):
        while True:
            msgs = await self._wire.recv()
            for msg in msgs:
                try:
                    self.process_message(msg)
                except:
                    logger.exception("Failed to process message %s", msg)

    def process_message(self, msg: Message):
        mt = msg.message_type
        if mt in [MessageType.METHOD_RETURN, MessageType.ERROR]:
            reply_to = msg.reply_serial
            if reply_to in self._futures:
                logger.info("Got response to %d", reply_to)
                f = self._futures.pop(reply_to)  # type: asyncio.Future
                if mt == MessageType.METHOD_RETURN:
                    f.set_result(msg.payload)
                else:
                    f.set_exception(DBusError(msg.payload))
                return
            else:
                logging.warning("Received unexpected response message: %s", msg)
        elif mt == MessageType.SIGNAL:
            self.process_signal(msg)
        elif mt == MessageType.METHOD_CALL:
            self.process_method_call(msg)
        else:
            logger.error("Don't know what to do with this: %s", msg)

    def process_signal(self, msg):
        logger.error("Received a signal: %s", msg)

    def process_method_call(self, msg):
        logger.error("Received a method call: %s, don't know what to do, sending NotImplemented error", msg)
        err = Message()
        err.message_type = MessageType.ERROR
        err.headers[HeaderField.DESTINATION] = msg.headers[HeaderField.SENDER]
        err.headers[HeaderField.REPLY_SERIAL] = debus.types.enforce_type(msg.serial, b'u')
        err.headers[HeaderField.ERROR_NAME] = 'space.equi.debus.Error.NotImplemented'
        self.send_message(err)

    def call(self, bus_name, object_path, interface_name, method, signature=None, args=None, timeout=None):
        # type: (str, str, str, str, str, typing.Any, float) -> typing.Any

        msg = make_mesage(MessageType.METHOD_CALL, bus_name, interface_name, method, object_path, signature, args)
        f = asyncio.Future()
        self._futures[msg.serial] = f
        logger.info("Sending method call: %s", msg)
        if timeout:
            def on_timeout():
                if not f.done():
                    f.set_exception(TimeoutError("Method call timed out: %s.%s" % (interface_name, method)))
            asyncio.get_event_loop().call_later(timeout, on_timeout)
        self.send_message(msg)
        return f

    def emit(self, object_path: str, interface_name: str, signal_name: str, signature=None, args=None):
        msg = Message()
        msg.message_type = MessageType.SIGNAL
        msg.interface = interface_name
        msg.path = object_path
        msg.member = signal_name
        if not signature is None:
            assert args
            msg.signature = signature
            msg.payload = args
        # print(msg)
        logger.warning("Going to send %s", msg)
        self.send_message(msg)

    def send_message(self, msg: Message):
        self._wire.send(msg)

    async def introspect(self, bus_name, object_path):
        logger.info("Introspecting %s %s", bus_name, object_path)
        result = await self.call(bus_name, object_path, 'org.freedesktop.DBus.Introspectable', 'Introspect', timeout=10)
        obj = IntrospectedObject(self, bus_name, object_path, result[0])
        return obj

    async def get_object_interface(self, bus_name, object_path, interface) -> IntrospectedInterface:
        return (await self.introspect(bus_name, object_path)).interfaces[interface]

    async def request_name(self, name: str):
        await self._freedesktop_interface.RequestName(name, 0)

    @property
    def freedesktop_interface(self):
        return self._freedesktop_interface


class ManagedConnection(ClientConnection):
    def __init__(self, uri):
        super().__init__(uri=uri)
        self._sub_mgr = debus.subscription.SubscriptionManager(self)
        self._obj_mgr = debus.objects.ObjectManager(self)

    def process_signal(self, msg):
        return self._sub_mgr.handle_message(msg)

    def process_method_call(self, msg):
        return self._obj_mgr.handle_call(msg)

    @property
    def sub_mgr(self):
        return self._sub_mgr

    @property
    def obj_mgr(self):
        return self._obj_mgr



from debus.subscription import MatchRule
from debus.connection import ClientConnection, ManagedConnection, DBusError
from debus.message import Message, MessageType, HeaderField
from debus.objects import DBusObject, DBusInterface
from debus.buses import SYSTEM, SESSION



from pathlib import Path


class DBusType:
    signature = None


class Signature(bytes, DBusType):
    signature = b'g'

    def __new__(cls, arg):
        if isinstance(arg, str):
            arg = arg.encode('utf-8')
        return super().__new__(cls, arg)

    __str__ = bytes.decode


class ObjectPath(bytes, DBusType):
    signature = b'o'

    def __new__(cls, arg):
        if isinstance(arg, str):
            arg = arg.encode('utf-8')
        if arg[0] != ord(b'/') or b'//' in arg:
            raise ValueError("Invalid object path: %s" % arg)
        return super().__new__(cls, arg)

    __str__ = bytes.decode

class enforce_type:
    """
    To be used as a function, like `call_with_variant_argument(enforce_type(b'u', 42))`, thus function-style name
    """
    def __init__(self, value, signature: bytes):
        assert isinstance(signature, bytes)
        self._value = value
        self._signature = signature

    @property
    def signature(self):
        return self._signature

    @property
    def value(self):
        return self._value

    def __repr__(self):
        return '<enforce_type[%s](%r)>' % (self.signature.decode(), self._value)


def guess_signature(arg):
    ret = None
    if isinstance(arg, DBusType):
        return arg.signature
    elif isinstance(arg, int):
        ret = b'i'
    elif isinstance(arg, float):
        ret = b'd'
    elif isinstance(arg, Path):
        ret = b'o'
    elif isinstance(arg, str):
        ret = b's'
    elif isinstance(arg, bytes):
        ret = b's'
    elif isinstance(arg, list):
        ret = b'a' + guess_signature(arg[0])
    elif isinstance(arg, tuple):
        ret = b'('
        for i in arg:
            ret += guess_signature(i)
        ret += b')'
    else:
        raise ValueError("Can't guess dbus type for %s" % (type(arg),))
    return ret



import sys
import os

if sys.platform.startswith('darwin'):
    SYSTEM = 'unix:path=/opt/local/var/run/dbus/system_bus_socket'
    SESSION = 'unix:path=%s' % (os.environ['DBUS_LAUNCHD_SESSION_BUS_SOCKET'],)
elif sys.platform.startswith('linux'):
    SYSTEM = 'unix:path=/var/run/dbus/system_bus_socket'
    SESSION = os.environ.get('DBUS_SESSION_BUS_ADDRESS', None)



import functools
from debus.message import Message, MessageType, HeaderField
import debus.connection

import logging
try:
    import typing
except:
    pass

logger = logging.getLogger(__name__)


def _equals_or_none(requirement, value):
    return requirement is None or requirement == value


@functools.total_ordering
class MatchRule:
    def __init__(self, msg_type:str='signal', sender:str=None, interface:str=None, member:str=None, path:str=None):
        self.msg_type = msg_type
        self.sender = sender
        self.interface = interface
        self.member = member
        self.path = path

    @property
    def rule_str(self):
        field2name = {
            'msg_type': 'type'
        }
        fields = {field2name.get(k, k): v for k,v in self.__dict__.items() if not v is None}
        return ','.join("%s='%s'" % (k, fields[k]) for k in sorted(fields))

    def match(self, msg: Message):
        return \
            _equals_or_none(self.msg_type, msg.message_type.name.lower()) and \
            _equals_or_none(self.sender, msg.headers[HeaderField.SENDER]) and \
            _equals_or_none(self.interface, msg.headers[HeaderField.INTERFACE]) and \
            _equals_or_none(self.member, msg.headers[HeaderField.MEMBER]) and \
            _equals_or_none(self.path, msg.headers[HeaderField.PATH])

    def __hash__(self):
        return self.rule_str.__hash__()

    def __eq__(self, other):
        if not isinstance(other, MatchRule): return False
        return self.rule_str == other.rule_str

    def __lt__(self, other):
        if not isinstance(other, MatchRule): return NotImplemented
        return self.rule_str.__lt__(other.rule_str)


class SubscriptionManager:
    def __init__(self, connection: debus.connection.ClientConnection):
        self._connection = connection
        self._matches = {}              # type: typing.Dict[MatchRule, list]

    def handle_message(self, m: debus.message.Message):
        # first build a list, then call, so user can subscribe/unsubscribe in message handler
        all_callbacks = []
        for rule, callbacks in self._matches.items():
            if rule.match(m):
                all_callbacks += callbacks
        for i in all_callbacks:
            i(m)
        return bool(all_callbacks)

    def subscribe(self, match: MatchRule, callback):
        if match in self._matches:
            cbl = self._matches[match]
        else:
            cbl = []
            self._matches[match] = cbl
            rule = match.rule_str
            logger.info("Subscribing with rule '%s'", rule)
            self._connection.freedesktop_interface.AddMatch(rule)
        cbl.append(callback)

    def unsubscribe(self, cb):
        for match in list(self._matches):
            v = self._matches[match]
            if cb in v:
                v.remove(cb)
            if not v:
                self._connection.freedesktop_interface.RemoveMatch(match.rule_str)
                del self._matches[match]




import asyncio
import binascii
import hashlib
import logging
import os
from pathlib import Path
import debus
import subprocess
import getpass
from debus.marshalling import InputBuffer, OutputBuffer
try:
    import typing
except:
    pass

logger = logging.getLogger(__name__)

class AbstractConnection:
    def __init__(self):
        self.reader: asyncio.StreamReader = None
        self.writer: asyncio.StreamWriter = None


class TcpConnection(AbstractConnection):
    def __init__(self, host, port):
        super().__init__()
        self.host = host
        self.port = port

    async def setup(self):
        logger.info('TCP socket connection %s:%s', self.host, self.port)
        self.reader, self.writer = await asyncio.open_connection(self.host, self.port)


class UnixConnection(AbstractConnection):
    def __init__(self, path):
        super().__init__()
        self.path = path

    async def setup(self):
        logger.info('Unix socket connection: %s', self.path)
        self.reader, self.writer = await asyncio.open_unix_connection(self.path)


class ExecConnection(AbstractConnection):

    def __init__(self, cmd, args):
        super().__init__()
        self.cmd = cmd
        self.args = args
        self.process = None

    async def setup(self):
        logger.info("Unix exec connection: %s %r", self.cmd, self.args)
        self.process = await asyncio.create_subprocess_exec(self.cmd, *self.args, stdout=subprocess.PIPE, stdin=subprocess.PIPE)
        self.reader = self.process.stdout
        self.writer = self.process.stdin

        pass

class WireConnection:
    def __init__(self, socket=None, uri=None):
        assert (socket is None) != (uri is None)
        self.reader = None          # type: asyncio.StreamReader
        self.writer = None          # type: asyncio.StreamWriter
        self.parser = InputBuffer()
        self.server_guid = None     # type: bytes
        if uri:
            if '%' in uri:
                raise NotImplementedError("Escape sequences ('%XX') in dbus address are not supported yet")
            kind, params = uri.split(':', maxsplit=1)
            params = {k:v for k,v in [i.split('=') for i in params.split(',')]}
            logger.info("Uri %r parsed, have to connect via %s with parameters %s", uri, kind, params)
            if kind == 'tcp':
                host, port = (params['host'], int(params['port']))
                self._connection = TcpConnection(host, port)
            elif kind == 'unix':
                socket_path = None
                if 'path' in params:
                    socket_path = params['path']
                elif 'abstract' in params:
                    socket_path = '\0' + params['abstract']
                else:
                    logger.error("Incorrect unix url: no 'path' or 'abstract' argument: %r", uri)
                    raise ValueError("Incorrect 'unix' uri provided")
                logger.warning("Will connect to unix socket %r", socket_path)
                self._connection = UnixConnection(socket_path)
            elif kind == 'unixexec':
                cmd = params['path']
                arg0 = params.get('argv0', cmd)
                args = []
                for i in range(1, 100):
                    key = 'argv%d' % i
                    if key in params:
                        args.append(params[key])
                    else:
                        break
                self._connection = ExecConnection(cmd, args)

    async def _check_auth_result(self):
        result = (await self.reader.readline())
        result_parts = result.split()
        status = result_parts[0]
        logger.warning("Auth result: %r", result)
        if status == b'OK':
            self.server_guid = binascii.unhexlify(result_parts[1])
            logger.warning("Auth ok, server guid: %s", self.server_guid)
            return True
        return False

    async def _do_auth_external(self, user_id: bytes):
        logger.warning("Attempting external auth (user %r)",  user_id.decode())
        self.writer.write(b"AUTH EXTERNAL %s\r\n" % binascii.hexlify(user_id))
        return await self._check_auth_result()

    async def _do_auth_anonymous(self):
        logger.warning("Attempting anonymous auth")
        self.writer.write(b"AUTH ANONYMOUS\r\n")
        return await self._check_auth_result()

    async def _do_auth_cookie(self):
        self.writer.write(b"AUTH DBUS_COOKIE_SHA1 %s\r\n" % binascii.hexlify(getpass.getuser().encode()))
        resp = await self.reader.readline()
        resp_parts = resp.split()
        if resp_parts[0] == b'REJECTED':
            logger.warning("Dbus cookie auth rejected")
            return False
        elif resp_parts[0] == b'DATA':
            s_cookie_context, s_cookie_id, s_challenge = binascii.unhexlify(resp_parts[1]).split()
            cookie_file = Path(os.path.expanduser('~/.dbus-keyrings')) / s_cookie_context.decode()
            logger.warning("Cookie auth data: %r", binascii.unhexlify(resp_parts[1]))
            selected_cookie = None
            for line in cookie_file.open('rb'):
                cookie_id, cookie_ts, cookie = line.split()
                if int(s_cookie_id) == int(cookie_id):
                    logging.warning("Found cookie %s", cookie)
                    selected_cookie = cookie
                    break

            assert not selected_cookie is None

            cli_challenge = b'a' * (len(s_challenge) // 2)
            response = b'%s:%s:%s' % (s_challenge, cli_challenge, selected_cookie)
            response_hash = hashlib.sha1(response).hexdigest().encode()

            response = binascii.hexlify(b'%s %s' % (cli_challenge, response_hash))
            self.writer.write(b'DATA %s\r\n' % response)
            logger.warning("Response to hash: %s, hash: %s, response: %s", response, response_hash, response)
            return await self._check_auth_result()
        else:
            raise RuntimeError("Unexpected response during DBUS_COOKIE_SHA1 auth: %r", resp)

    async def connect_and_auth(self):
        await self._connection.setup()
        self.reader, self.writer = self._connection.reader, self._connection.writer
        self.writer.write(b"\0")
        self.writer.write(b"AUTH\r\n")
        response = await self.reader.readline()
        auth_available = response.split()[1:]
        logger.warning("Auth types available: %s", auth_available)
        external_id_override = os.environ.get('DEBUS_EXTERNAL_USERID', None)
        if False:
            pass
        elif await self._do_auth_cookie():
            pass
        elif not external_id_override is None and await self._do_auth_external(external_id_override.encode()):
            pass
        elif await self._do_auth_external(b'%d' % os.getuid()):
            pass
        elif await self._do_auth_external(os.environ['USER'].encode()):
            pass
        elif await self._do_auth_anonymous():
            pass
        else:
            logger.error("Can't authenticate with the server")
        self.writer.write(b'BEGIN\r\n')

    async def recv(self) -> 'typing.List[debus.Message]':
        msgs = []
        while len(msgs) == 0:
            data = await self.reader.read(1024)
            if len(data) == 0:
                logger.error("Connection closed, buffer content: %s", self.parser)
                raise RuntimeError("Connection closed")
            msgs = self.parser.feed_data(data)
        return msgs

    def send(self, message: 'debus.Message'):
        buf = OutputBuffer()
        try:
            buf.put_message(message)
        except:
            logger.error("Unable to serialize message %s", message)
            raise
        data = buf.get()
        assert len(InputBuffer().feed_data(data)) == 1
        self.writer.write(data)



import setuptools
import os
import sys
import setuptools.command.build_py
import typing
from pathlib import Path

def maybe_cythonize(exts: typing.List[setuptools.Extension]):
    do_it = False
    for ext in exts:
        for i in ext.sources:
            c = Path(i)
            pyx = Path(i.replace('.c', '.pyx'))
            if pyx.exists() and ((not c.exists()) or pyx.stat().st_mtime > c.stat().st_mtime):
                ext.sources.remove(i)
                ext.sources.append(str(pyx))
                do_it = True

    if do_it:
        from Cython.Build import cythonize
        exts = cythonize(exts)
    return exts


class DeannotateAndBuild(setuptools.command.build_py.build_py):

    def build_packages(self):
        super().build_packages()
        if sys.version_info[0] == 3 and sys.version_info[1] < 6:
            print("Installing for older python, trying to clean up annotations...", file=sys.stderr)
            if sys.platform.startswith('darwin'):
                sed_i = 'sed -i""'
            else:
                sed_i = 'sed -i'
            os.system(r'''set -x; cd %s; find . -name '*.py' | xargs -n1 %s 's@^\( *[a-zA-Z._]*\):.*=@\1 = @g' ''' % (self.build_lib, sed_i))


setuptools.setup(
    cmdclass={
        'build_py': DeannotateAndBuild,
    },
    name='debus',
    version='0.1.11',
    description='DBus wire protocol implementation',
    long_description=Path('README.md').read_text(),
    long_description_content_type="text/markdown",
    author='Vladimir Shapranov',
    author_email='equidamoid@gmail.com',
    url='https://github.com/Equidamoid/debus',
    ext_modules=maybe_cythonize([setuptools.Extension('debus.marshalling', ["debus/marshalling.c"])]),
    packages=setuptools.find_packages(),
    install_requires=[
        'lxml',
    ],
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',

    ]

)



from debus.objects import DBusObject, DBusInterface, dbus_method, MethodInfo, PropertyInfo, SignalInfo
from lxml import etree
from debus.marshalling import split_signature
import logging

logger = logging.getLogger(__name__)

class IntrospectInterface(DBusInterface):
    name = 'org.freedesktop.DBus.Introspectable'

    def __init__(self):
        super().__init__()
        self.child_objects = []

    @dbus_method('', 's')
    def Introspect(self):
        root = etree.Element('node', name=self._obj.path)
        for iface in self._obj.interfaces:
            iface_node = etree.Element('interface', name=iface.name)
            for m in iface.methods:
                m_info: MethodInfo = m._pybus_method

                method_node = etree.Element('method', name=m_info.name)
                in_args = split_signature(m_info.in_signature.encode())
                out_args = split_signature(m_info.out_signature.encode())
                for i in in_args:
                    arg_node = etree.Element('arg', type=i, direction='in')
                    method_node.append(arg_node)
                for i in out_args:
                    arg_node = etree.Element('arg', type=i, direction='out')
                    method_node.append(arg_node)
                iface_node.append(method_node)
            root.append(iface_node)
            for m in iface.signals:
                s_info: SignalInfo = m._pybus_signal
                signal_node = etree.Element('signal', name=s_info.name)
                out_args = split_signature(s_info.signature.encode())
                out_arg_names = s_info.argument_names
                if len(out_args) != len(out_arg_names):
                    logger.error("Argument names don't match signature: %s vs %s", out_arg_names, out_args)
                for n, i in enumerate(out_args):
                    arg_node = etree.Element('arg', type=i, name=out_arg_names[n], direction='out')
                    signal_node.append(arg_node)
                iface_node.append(signal_node)
            root.append(iface_node)
        for i in self.child_objects:
            root.append(etree.Element('node', name=i))
        return etree.tostring(root, pretty_print=True),
